pub struct ProgramInfoBuilder {
    addr: u32,
    buf: [u8; 4096],
    info_pos: usize,
    data_pos: usize,
    info_table_size: usize,
    data_end_offset: usize,
}

pub struct ProgramInfoBuilderForGroup {
    builder: ProgramInfoBuilder,
    tag: [char; 2],
    id: u32,
}

#[repr(u16)]
#[allow(non_camel_case_types)]
#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum GroupFlags {
    NONE            = 0x0000,
    SHOW_IF_EMPTY   = 0x0001,  // default is to hide
    SEPARATE_COMMAS = 0x0002,  // default is newlines
    SORT_ALPHA      = 0x0004,  // default is no sort
    ADVANCED        = 0x0008,  // if set, then only shown in say info -a
}

#[repr(u8)]
#[allow(non_camel_case_types)]
#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum PinFunction {
    XIP = 0,
    SPI = 1,
    UART = 2,
    I2C = 3,
    PWM = 4,
    SIO = 5,
    PIO0 = 6,
    PIO1 = 7,
    GPCK = 8,
    USB = 9,
    NULL = 0xf,
}

const fn copy(mut buf: [u8; 4096], offset: usize, source: &[u8], source_offset: usize) -> [u8; 4096] {
    if source_offset >= source.len() {
        buf
    } else {
        buf[offset] = source[source_offset];
        copy(buf, offset+1, source, source_offset+1)
    }
}

const fn copy2<const N: usize, const N2: usize>(mut acc: [u8; N], source: &[u8], index: usize) -> [u8; N] {
    if index >= N2 {
        acc
    } else {
        acc[index] = source[index];
        copy2::<N, N2>(acc, source, index+1)
    }
}
const fn copy3<const N: usize, const N2: usize>(source: &[u8]) -> [u8; N] {
    copy2::<N, N2>([0; N], source, 0)
}

impl ProgramInfoBuilder {
    pub const fn new(addr: u32, info_table_size: u32, copy_table_size: u32) -> Self {
        let copy_table_size = copy_table_size as usize;

        core::assert!(info_table_size % 4 == 0);

        ProgramInfoBuilder {
            addr,
            buf: {
                let mut x = [0xff; 4096];
                // mark end of copy table with zero in first entry
                x[4096 - copy_table_size + 0] = 0;
                x[4096 - copy_table_size + 1] = 0;
                x[4096 - copy_table_size + 2] = 0;
                x[4096 - copy_table_size + 3] = 0;
                x
            },
            info_pos: 0,
            data_pos: info_table_size as usize,
            data_end_offset: 4096 - copy_table_size as usize,
            info_table_size: info_table_size as usize,
        }
    }

    pub const fn build(mut self: Self) -> [u8; 4096] {
        if self.info_pos < self.info_table_size {
            self = self.align();
            let addr = self.addr + self.data_pos as u32;
            self = self.push_u32(0xffffffff).fill_info_table(addr);
        }
        self.buf
    }

    const fn push_info_u8(mut self: Self, value: u8) -> Self {
        self.buf[self.info_pos] = value;
        self.info_pos += 1;
        if self.info_pos > self.info_table_size {
            core::panic!("ProgramInfoBuilder: too much data for info table")
        }
        self
    }

    const fn push_info_ptr(self: Self, value: u32) -> Self {
        self
            .push_info_u8((value & 0xff) as u8)
            .push_info_u8(((value >> 8) & 0xff) as u8)
            .push_info_u8(((value >> 16) & 0xff) as u8)
            .push_info_u8(((value >> 24) & 0xff) as u8)
    }

    const fn push_current_info_ptr(mut self: Self) -> Self {
        self = self.align();
        let addr = self.addr + self.data_pos as u32;
        self.push_info_ptr(addr)
    }

    const fn fill_info_table(mut self: Self, value: u32) -> Self {
        if self.info_pos < self.info_table_size {
            self = self.push_info_ptr(value);
        }
        if self.info_pos < self.info_table_size {
            self = self.push_info_ptr(value);
        }
        if self.info_pos < self.info_table_size {
            self = self.push_info_ptr(value);
        }
        if self.info_pos < self.info_table_size {
            self = self.push_info_ptr(value);
        }
        if self.info_pos < self.info_table_size {
            self.fill_info_table(value)
        } else {
            self
        }
    }

    const fn push(mut self: Self, value: u8) -> Self {
        self.buf[self.data_pos] = value;
        self.data_pos += 1;
        if self.data_pos > self.data_end_offset {
            core::panic!("ProgramInfoBuilder: too much data")
        }
        self
    }

    pub const fn align(mut self: Self) -> Self {
        if self.data_pos % 4 != 0 {
            self = self.push(0xff);
        }
        if self.data_pos % 4 != 0 {
            self = self.push(0xff);
        }
        if self.data_pos % 4 != 0 {
            self = self.push(0xff);
        }
        self
    }

    const fn push_u16(self: Self, value: u16) -> Self {
        self
            .push((value & 0xff) as u8)
            .push(((value >> 8) & 0xff) as u8)
    }

    const fn push_u32(self: Self, value: u32) -> Self {
        self
            .push((value & 0xff) as u8)
            .push(((value >> 8) & 0xff) as u8)
            .push(((value >> 16) & 0xff) as u8)
            .push(((value >> 24) & 0xff) as u8)
    }

    const fn push_tag(self: Self, tag: [char; 2]) -> Self {
        self.push(tag[0] as u8)
            .push(tag[1] as u8)
    }

    const fn push_extra_data(mut self: Self, value: &[u8]) -> (Self, u32) {
        if self.data_end_offset - self.data_pos < value.len() {
            core::panic!("ProgramInfoBuilder: too much data")
        }
        let offset = self.data_pos;
        //self.buf[offset..offset+value.len()].copy_from_slice(value);
        self.buf = copy(self.buf, offset, value, 0);

        self.data_pos += value.len();
        let addr = self.addr + offset as u32;
        (self, addr)
    }

    //NOTE This will push the address and then add the string so this must be the last item of the data record.
    const fn push_string_utf8(mut self: Self, value: &[u8]) -> Self {
        if self.data_pos - self.info_pos < value.len() + 1 {
            core::panic!("ProgramInfoBuilder: too much data")
        }

        // push dummy value (will be replaced by address)
        let x1 = self.data_pos;
        self = self.push_u32(0);

        let (self2, addr) = self.push_extra_data(value);
        self = self2;
        (self, _) = self.push_extra_data(&[0]);

        // replace dummy by actual address
        let x2 = self.data_pos;
        self.data_pos = x1;
        self = self.push_u32(addr);
        self.data_pos = x2;
        
        self
    }

    //NOTE This will push the address and then add the string so this must be the last item of the data record.
    const fn push_string(self: Self, value: &str) -> Self {
        self.push_string_utf8(value.as_bytes())
    }

    const fn append_to_last_string(mut self: Self, value: &str) -> Self {
        if self.data_pos - self.info_pos < value.len() {
            core::panic!("ProgramInfoBuilder: too much data")
        }

        self.data_pos -= 1;
        (self, _) = self.push_extra_data(value.as_bytes());
        (self, _) = self.push_extra_data(&[0]);
        
        self
    }

    const fn bi_int(self: Self, tag: [char; 2], id: u32, value: u32) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_ID_AND_INT)
            .push_tag(tag)
            .push_u32(id)
            .push_u32(value)
    }

    const fn bi_string(self: Self, tag: [char; 2], id: u32, value: &str) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_ID_AND_STRING)
            .push_tag(tag)
            .push_u32(id)
            .push_string(value)
    }

    const fn bi_string_utf8(self: Self, tag: [char; 2], id: u32, value: &[u8]) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_ID_AND_STRING)
            .push_tag(tag)
            .push_u32(id)
            .push_string_utf8(value)
    }

    const fn bi_named_group(self: Self, tag: [char; 2], id: u32, flags: u16, group_tag: [char; 2], group_id: u32, label: &str) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_NAMED_GROUP)
            .push_tag(tag)
            .push_u32(id)
            .push_u16(flags)
            .push_tag(group_tag)
            .push_u32(group_id)
            .push_string(label)
    }

    const fn bi_encoded_pins_with_func(self: Self, value: u32) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_PINS_WITH_FUNC)
            .push_tag(Self::BINARY_INFO_TAG_RASPBERRY_PI)
            .push_u32(value)
    }

    const fn bi_pins_with_names(self: Self, pin_mask: u32, label: &str) -> Self {
        self.push_current_info_ptr()
            .push_u16(Self::BINARY_INFO_TYPE_PINS_WITH_NAME)
            .push_tag(Self::BINARY_INFO_TAG_RASPBERRY_PI)
            .push_u32(pin_mask)
            .push_string(label)
    }
    
    const BINARY_INFO_TAG_RASPBERRY_PI: [char; 2] = ['R', 'P'];

    #[allow(unused)] const BINARY_INFO_TYPE_RAW_DATA: u16 = 1;
    #[allow(unused)] const BINARY_INFO_TYPE_SIZED_DATA: u16 = 2;
    #[allow(unused)] const BINARY_INFO_TYPE_BINARY_INFO_LIST_ZERO_TERMINATED: u16 = 3;
    #[allow(unused)] const BINARY_INFO_TYPE_BSON: u16 = 4;
    #[allow(unused)] const BINARY_INFO_TYPE_ID_AND_INT: u16 = 5;
    #[allow(unused)] const BINARY_INFO_TYPE_ID_AND_STRING: u16 = 6;
    #[allow(unused)] const BINARY_INFO_TYPE_BLOCK_DEVICE: u16 = 7;
    #[allow(unused)] const BINARY_INFO_TYPE_PINS_WITH_FUNC: u16 = 8;
    #[allow(unused)] const BINARY_INFO_TYPE_PINS_WITH_NAME: u16 = 9;
    #[allow(unused)] const BINARY_INFO_TYPE_PINS_WITH_NAMES: u16 = 9;
    #[allow(unused)] const BINARY_INFO_TYPE_NAMED_GROUP: u16 = 10;

    const BINARY_INFO_ID_RP_PROGRAM_NAME: u32 = 0x02031c86;
    const BINARY_INFO_ID_RP_PROGRAM_VERSION_STRING: u32 = 0x11a9bc3a;
    const BINARY_INFO_ID_RP_PROGRAM_BUILD_DATE_STRING: u32 = 0x9da22254;
    const BINARY_INFO_ID_RP_BINARY_END: u32 = 0x68f465de;
    const BINARY_INFO_ID_RP_PROGRAM_URL: u32 = 0x1856239a;
    const BINARY_INFO_ID_RP_PROGRAM_DESCRIPTION: u32 = 0xb6a07c19;
    const BINARY_INFO_ID_RP_PROGRAM_FEATURE: u32 = 0xa1f4b453;
    const BINARY_INFO_ID_RP_PROGRAM_BUILD_ATTRIBUTE: u32 = 0x4275f0d3;
    const BINARY_INFO_ID_RP_SDK_VERSION: u32 = 0x5360b3ab;
    const BINARY_INFO_ID_RP_PICO_BOARD: u32 = 0xb63cffbb;
    const BINARY_INFO_ID_RP_BOOT2_NAME: u32 = 0x7f8882e1;

    pub const fn binary_end_address(self: Self, value: u32) -> Self {
        self.bi_int(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_BINARY_END, value)
    }

    pub const fn program_name(self: Self, name: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_NAME, name)
    }
    pub const fn program_description(self: Self, description: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_DESCRIPTION, description)
    }
    pub const fn program_version_string(self: Self, version_string: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_VERSION_STRING, version_string)
    }
    pub const fn program_build_date_string(self: Self, date_string: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_BUILD_DATE_STRING, date_string)
    }
    pub const fn program_url(self: Self, url: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_URL, url)
    }
    // multiple of these may be added
    pub const fn program_feature(self: Self, feature: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_FEATURE, feature)
    }
    pub const fn program_build_attribute(self: Self, attr: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_BUILD_ATTRIBUTE, attr)
    }
    pub const fn sdk_version(self: Self, attr: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_SDK_VERSION, attr)
    }
    pub const fn board(self: Self, attr: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PICO_BOARD, attr)
    }
    pub const fn boot2_name(self: Self, attr: &str) -> Self {
        self.bi_string(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_BOOT2_NAME, attr)
    }
    pub const fn program_feature_group(self: Self, tag: [char; 2], id: u32, label: &str) -> ProgramInfoBuilderForGroup {
        self.program_feature_group_with_flags(tag, id, label, GroupFlags::NONE)
    }
    pub const fn program_feature_group_with_flags(self: Self, tag: [char; 2], id: u32, label: &str, flags: GroupFlags) -> ProgramInfoBuilderForGroup {
        let builder = self.bi_named_group(Self::BINARY_INFO_TAG_RASPBERRY_PI, Self::BINARY_INFO_ID_RP_PROGRAM_FEATURE,
            flags as u16, tag, id, label);
        ProgramInfoBuilderForGroup { builder, tag, id }
    }

    pub const fn pins_with_func(self: Self, func: PinFunction, pins: &[u8]) -> Self {
        assert!((func as u8) < 16);
        assert!(pins.len() > 0 && pins.len() < 6);

        const BI_PINS_ENCODING_MULTI: u32 = 2;
        let encoded = if pins.len() == 1 {
            BI_PINS_ENCODING_MULTI | ((func as u32) << 3) | ((pins[0] as u32) << 7) | ((pins[0] as u32) << 12)
        } else if pins.len() == 2 {
            BI_PINS_ENCODING_MULTI | ((func as u32) << 3) | ((pins[0] as u32) << 7) | ((pins[1] as u32) << 12) | ((pins[1] as u32) << 17)
        } else if pins.len() == 3 {
            BI_PINS_ENCODING_MULTI | ((func as u32) << 3) | ((pins[0] as u32) << 7) | ((pins[1] as u32) << 12) | ((pins[2] as u32) << 17) | ((pins[2] as u32) << 22)
        } else if pins.len() == 4 {
            BI_PINS_ENCODING_MULTI | ((func as u32) << 3) | ((pins[0] as u32) << 7) | ((pins[1] as u32) << 12) | ((pins[2] as u32) << 17) | ((pins[3] as u32) << 22) | ((pins[3] as u32) << 27)
        } else if pins.len() == 5 {
            BI_PINS_ENCODING_MULTI | ((func as u32) << 3) | ((pins[0] as u32) << 7) | ((pins[1] as u32) << 12) | ((pins[2] as u32) << 17) | ((pins[3] as u32) << 22) | ((pins[4] as u32) << 27)
        } else {
            assert!(false);
            0
        };

        self.bi_encoded_pins_with_func(encoded)
    }

    pub const fn pin_range_with_func(self: Self, func: PinFunction, pin_range: (u8, u8)) -> Self {
        assert!((func as u8) < 16);
        let (plo, phi) = pin_range;

        const BI_PINS_ENCODING_RANGE: u32 = 1;
        let encoded = BI_PINS_ENCODING_RANGE | ((func as u32) << 3) | ((plo as u32) << 7) | ((phi as u32) << 12);

        self.bi_encoded_pins_with_func(encoded)
    }

    pub const fn pins_with_names(mut self: Self, pins: &[(u8, &str)]) -> Self {
        const fn collect_mask(mask: u32, pins: &[(u8, &str)], offset: usize) -> u32 {
            if offset >= pins.len() {
                mask
            } else {
                assert!(offset == 0 || pins[offset-1].0 < pins[offset].0, "Pins must be in ascending order");
                collect_mask(mask | (1 << pins[offset].0), pins, offset+1)
            }
        }
        self = self.bi_pins_with_names(collect_mask(0, pins, 0), pins[0].1);

        const fn append_names(acc: ProgramInfoBuilder, pins: &[(u8, &str)], offset: usize) -> ProgramInfoBuilder {
            if offset >= pins.len() {
                acc
            } else {
                append_names(acc.append_to_last_string("|").append_to_last_string(pins[offset].1), pins, offset+1)
            }
        }
        append_names(self, pins, 1)
    }
}

impl ProgramInfoBuilderForGroup {
    pub const fn feature(mut self: Self, name: &str) -> Self {
        self.builder = self.builder.bi_string(self.tag, self.id, name);
        self
    }
    pub const fn feature_utf8(mut self: Self, name: &[u8]) -> Self {
        self.builder = self.builder.bi_string_utf8(self.tag, self.id, name);
        self
    }
    pub const fn feature_utf8_part<const N: usize>(mut self: Self, name: &[u8]) -> Self {
        let name: [u8; N] = copy3::<N, N>(name);
        self.builder = self.builder.bi_string_utf8(self.tag, self.id, &name);
        self
    }
    pub const fn group_done(self: Self) -> ProgramInfoBuilder {
        self.builder
    }
}
